import bleach
import json
import inspect
import markdown
import traceback
import pyparsing

from django.apps import apps
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import (
    PermissionRequiredMixin,
    UserPassesTestMixin,
)
from django.core import serializers
from django.core.exceptions import FieldError, ObjectDoesNotExist
from django.db import models
from django.db import transaction
from django.db.models import Count
from django.http.response import JsonResponse, Http404
from django.urls import reverse
from django.urls.base import reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from django.views.generic.base import TemplateView
from django.views.generic.edit import (
    UpdateView,
    CreateView,
    ModelFormMixin,
    DeleteView,
)
from django.views.generic.list import ListView
from reversion.views import RevisionMixin

import massmailer.forms
import massmailer.models
import massmailer.tasks
from massmailer.query_parser import QueryParser
from massmailer.utils import JinjaEscapeExtension


class MailerAdminMixin(UserPassesTestMixin):
    def test_func(self):
        # Like for the Django Admin, only staff users can see the mailing
        # panel, whatever their permissions are.
        return self.request.user.is_staff


class ObjectByIdMixin:
    def get_object(self, queryset=None):
        try:
            return (queryset or self.get_queryset()).get(pk=self.kwargs['id'])
        except ObjectDoesNotExist:
            raise Http404()


class DashboardView(MailerAdminMixin, TemplateView):
    template_name = 'massmailer/dashboard.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['templates'] = massmailer.models.Template.objects.annotate(
            query_count=Count('useful_queries')
        )
        context['queries'] = massmailer.models.Query.objects.annotate(
            template_count=Count('useful_with')
        )
        return context


class TemplateMixin(MailerAdminMixin, RevisionMixin):
    template_name = 'massmailer/template/details.html'
    context_object_name = 'template'
    model = massmailer.models.Template
    form_class = massmailer.forms.TemplateForm


class CreateTemplateView(PermissionRequiredMixin, TemplateMixin, CreateView):
    permission_required = 'massmailer.create_template'

    def get_success_url(self):
        return self.object.get_absolute_url()

    def form_valid(self, form):
        self.object = template = form.save(commit=False)
        template.author = self.request.user
        template.save()
        return super(ModelFormMixin, self).form_valid(form)

    def get_object(self):
        return None  # the template isn't in db yet


class UpdateTemplateView(
    PermissionRequiredMixin, TemplateMixin, ObjectByIdMixin, UpdateView
):
    permission_required = 'massmailer.change_template'

    def form_valid(self, form):
        if form.was_overwritten():
            form.reset_overwritten()
            form.add_error(
                None,
                _(
                    "The form was modified and saved while you were editing. "
                    "Please merge with current version."
                ),
            )
            return self.form_invalid(form)
        return super().form_valid(form)


@method_decorator(csrf_exempt, name='dispatch')
class TemplatePreviewView(PermissionRequiredMixin, MailerAdminMixin, View):
    permission_required = 'massmailer.view_template'

    def post(self, request, *args, **kwargs):
        # TODO: context from query builder
        data = {}

        if not request.user.has_perm('massmailer.view_query'):
            data['error'] = _(
                "You don't have the permission to view query results."
            )
            return JsonResponse(data)

        html_enabled = request.POST.get('html_enabled') == 'true'

        query = massmailer.models.Query.objects.get(pk=(request.POST['query']))
        page = int(request.POST['page'])

        template = massmailer.models.Template()
        template.subject = request.POST['subject']
        template.plain_body = request.POST['plain']
        template.language = request.POST['language']

        if html_enabled:
            if request.POST.get('use_markdown') == 'true':
                md = markdown.Markdown(extensions=[JinjaEscapeExtension()])
                html = data['html_template'] = md.convert(
                    bleach.clean(template.plain_body)
                )
            else:
                html = request.POST['html']
            template.html_body = html

        results, user_qs = query.get_results()
        qs = results.queryset
        count = qs.count()
        user_count = user_qs.count()
        data['query'] = {
            'count': count,
            'user_count': user_count,
            'page': min(count, page),
        }
        try:
            object = qs[page]
            context = {
                alias: getattr(object, field)
                for alias, field in results.aliases.items()
            }
            context[results.model_name] = object
            data['render'] = template.full_preview(context)
        except IndexError:
            data['render'] = ''
        return JsonResponse(data)


class QueryMixin(MailerAdminMixin, RevisionMixin):
    template_name = 'massmailer/query/details.html'
    context_object_name = 'query'
    model = massmailer.models.Query
    form_class = massmailer.forms.QueryForm

    @cached_property
    def available_enums(self):
        return sorted(
            (
                {'name': name, 'members': [m.name for m in enum]}
                for name, enum in QueryParser().available_enums.items()
            ),
            key=lambda e: e['name'].lower(),
        )

    @cached_property
    def available_funcs(self):
        return [
            {
                'name': name,
                'doc': inspect.getdoc(func),
                'signature': inspect.signature(func),
            }
            for name, func in QueryParser().available_funcs.items()
        ]

    @cached_property
    def available_models(self):
        User = get_user_model()

        def find_user_field(model):
            for field in model._meta.fields:
                try:
                    if field.related_model is User:
                        return field.name
                except AttributeError:
                    pass

        models = apps.get_models()
        models = (
            {
                'name': str(model._meta.verbose_name_plural.capitalize()),
                'cls_name': model.__name__,
                'app': model._meta.app_label,
                'is_user': model is User,
                'user_field': find_user_field(model),
                'label': model._meta.label,
                'doc': inspect.getdoc(model),
            }
            for model in models
        )
        return sorted(
            models,
            key=lambda e: (
                not e['is_user'],
                e['user_field'] is None,
                e['app'].lower(),
                e['name'].lower(),
            ),
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['available_enums'] = self.available_enums
        context['available_funcs'] = self.available_funcs
        context['available_models'] = self.available_models
        context['user_model'] = get_user_model().__name__
        return context


class CreateQueryView(PermissionRequiredMixin, QueryMixin, CreateView):
    permission_required = 'massmailer.create_query'

    def get_success_url(self):
        return self.object.get_absolute_url()

    def form_valid(self, form):
        self.object = query = form.save(commit=False)
        query.author = self.request.user
        query.save()
        return super(ModelFormMixin, self).form_valid(form)

    def get_object(self):
        return None  # the query isn't in the db yet


class UpdateQueryView(
    PermissionRequiredMixin, QueryMixin, ObjectByIdMixin, UpdateView
):
    permission_required = 'massmailer.change_query'


@method_decorator(csrf_exempt, name='dispatch')
class QueryPreviewView(PermissionRequiredMixin, MailerAdminMixin, View):
    permission_required = 'massmailer.view_query'

    @staticmethod
    def serialize(obj):
        if obj is None:
            return None
        if not isinstance(obj, models.Model):
            return None
        return json.loads(
            serializers.get_serializer('json')().serialize([obj])
        )[0]

    def post(self, request, *args, **kwargs):
        query = request.POST['query']
        page = int(request.POST['page'])
        try:
            # run the query
            result, user_qs = massmailer.models.Query.execute(query)
            qs = result.queryset
            count = qs.count()
            user_count = user_qs.count()
            instance = None
            if 0 <= page < count:
                obj = qs[page]
                instance = {result.model_name: self.serialize(obj)}
                for name, field in result.aliases.items():
                    serialized = self.serialize(getattr(obj, field))
                    if serialized:
                        instance[name] = serialized

            data = {
                'count': count,
                'user_count': user_count,
                'model': qs.model._meta.label,
                'model_name': result.model_name,
                'aliases': list(result.aliases.items()),
                'query': str(qs.query),
                'result': instance,
            }
        except Exception as e:
            if isinstance(e, (massmailer.query_parser.ParseError, FieldError)):
                error = str(e)
            elif isinstance(e, pyparsing.ParseException):
                error = _("Syntax error at position %(pos)s.") % {'pos': e.loc}
            else:
                error = traceback.format_exc(limit=2)
            data = {'error': error}
        return JsonResponse(data)


class BatchListView(PermissionRequiredMixin, MailerAdminMixin, ListView):
    model = massmailer.models.Batch
    template_name = 'massmailer/batch-list.html'
    context_object_name = 'batches'
    paginate_by = 25
    permission_required = 'massmailer.view_batch'


class BatchCreateView(PermissionRequiredMixin, MailerAdminMixin, CreateView):
    model = massmailer.models.Batch
    form_class = massmailer.forms.CreateBatchForm
    template_name = 'massmailer/batch-create.html'
    permission_required = 'massmailer.create_batch'

    def get_success_url(self):
        return reverse('massmailer:batch:detail', args=[self.object.pk])

    def form_valid(self, form):
        with transaction.atomic():
            # create the batch
            batch = form.save(commit=False)
            batch.initiator = self.request.user
            batch.save()

            emails = list(batch.build_emails())
            # create the batch emails
            massmailer.models.BatchEmail.objects.bulk_create(emails)

        # create the tasks
        for email in emails:
            email.send_task()

        return super().form_valid(form)

    def get_object(self):
        return None  # the batch isn't in the db yet


class BatchDetailView(PermissionRequiredMixin, MailerAdminMixin, ListView):
    model = massmailer.models.BatchEmail
    template_name = 'massmailer/batch-emails.html'
    context_object_name = 'emails'
    paginate_by = 200
    permission_required = 'massmailer.view_batch'

    @property
    def batch_id(self):
        return self.kwargs['id']

    @cached_property
    def batch(self):
        return (
            massmailer.models.Batch.objects.prefetch_related('emails')
            .annotate(email_count=Count('emails'))
            .get(pk=self.batch_id)
        )

    def get_queryset(self):
        return self.batch.emails.all()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['batch'] = self.batch
        return context


class BatchRetryView(PermissionRequiredMixin, MailerAdminMixin, UpdateView):
    model = massmailer.models.Batch
    pk_url_kwarg = 'id'
    fields = []
    success_url = reverse_lazy('massmailer:batch:list')
    permission_required = 'massmailer.change_batch'

    def form_valid(self, form):
        # create the tasks
        for email in self.get_object().pending_emails():
            email.send_task()
        return super(ModelFormMixin, self).form_valid(form)


class BatchDeleteView(PermissionRequiredMixin, MailerAdminMixin, DeleteView):
    model = massmailer.models.Batch
    pk_url_kwarg = 'id'
    success_url = reverse_lazy('massmailer:batch:list')
    permission_required = 'massmailer.delete_batch'



import ast
import re
import pyparsing as p

from django.apps import apps
from django.db import models
from django.db.models import Q


class ParseError(ValueError):
    pass


class CaseInsensitive(str):
    pass


class ParseResult:
    queryset = None
    model_name = None
    aliases = {}


def _find_subclasses(cls):
    found = {}
    subclasses = cls.__subclasses__()
    if subclasses:
        for sub_cls in subclasses:
            found.update(_find_subclasses(sub_cls))
    elif not cls.__name__.startswith('_'):
        found[cls.__name__.lower()] = cls
    return found


class QueryParser:
    """
    A parser for a simple grammar to write user-friendly queries that get
    translated to Django ORM.

    Aggregate functions such as count() shall be registered in the
    available_funcs mapping:
        func name → func
    Use load_django_funcs=True to load all known Django ORM functions.

    Available Django models shall be registered in the available_models
    mapping:
        model name → model class
    Use load_django_models=True to load all known Django models for the current
    project.

    There also is a concept of literal enums that have to be registered in the
    available_enums mapping:
        (enum name eg. 'Namespace.EnumName') → enum class
    where 'enum class' shall implement Python's enum.Enum API, ie.
    EnumCls[name] and EnumCls(value) methods.

    Syntax example:

        # this is a comment
        SomeModel [as name]
          .field = 42
          # two field predicates are joined with "and" if not explicitly using
          # "or"
          count(.related_field) > 10
          # there is a support for literal enums, that get replaced with their
          # value
          .field = SomeClass.SomeEnum.some_enum_member
          (.field contains "string" or
           .field contains i"case insensitive")

        alias some_name = .some_field
    """

    def __init__(
        self, load_django_funcs=True, load_django_models=True, load_enums=True
    ):
        self.available_funcs = {}
        self.available_models = {}
        self.available_enums = {}

        if load_django_funcs:
            self.available_funcs.update(_find_subclasses(models.Func))

        if load_django_models:
            self.available_models.update(
                {model.__name__: model for model in apps.get_models()}
            )

        if load_enums:
            from massmailer import REGISTERED_ENUMS

            self.available_enums.update(REGISTERED_ENUMS)

    def parse_query(self, query: str) -> ParseResult:
        func_i = 0

        def parse_enum(tokens):
            enum, member = tokens[0].rsplit('.', 1)
            try:
                enum = self.available_enums[enum]
            except KeyError:
                raise ParseError(
                    "Unknown enum '{}'.\n\nAvailable enums: {}".format(
                        enum, ', '.join(self.available_enums.keys())
                    )
                )
            try:
                # try with member name
                member = enum[member]
            except KeyError:
                # try with member value
                try:
                    member = enum(member)
                except ValueError:
                    raise ParseError(
                        "invalid member '{}' of {}".format(member, enum)
                    )
            return member.value

        def parse_string(tokens):
            string = tokens['string']
            if tokens.get('nocase'):
                return CaseInsensitive(string)
            return string

        def parse_field(tokens):
            return '__'.join(tokens)

        def parse_clause(tokens):
            nonlocal func_i
            annotations = {}
            field = tokens.get('field')
            value = tokens.get('value')
            negation = tokens.get('negation', False)
            insensitive = isinstance(value, CaseInsensitive)

            if tokens.get('operation'):
                func = self.available_funcs[tokens['func_name']]
                name = 'func_{}'.format(func_i)
                func_i += 1
                annotations[name] = func(tokens['func_field'])
                field = name

            if tokens.get('='):
                if insensitive:
                    field += '__iexact'
            elif tokens.get('!='):
                negation = True
            elif tokens.get('<='):
                field += '__lte'
            elif tokens.get('>='):
                field += '__gte'
            elif tokens.get('<'):
                field += '__lt'
            elif tokens.get('>'):
                field += '__gt'
            elif tokens.get('empty'):
                value = ''
            elif tokens.get('null'):
                field += '__isnull'
                value = not negation
                negation = False
            elif tokens.get('contains'):
                if insensitive:
                    field += '__icontains'
                else:
                    field += '__contains'
            elif tokens.get('startswith'):
                if insensitive:
                    field += '__istartswith'
                else:
                    field += '__startswith'
            elif tokens.get('endswith'):
                if insensitive:
                    field += '__iendswith'
                else:
                    field += '__endswith'
            elif tokens.get('matches'):
                if insensitive:
                    field += '__iregex'
                else:
                    field += '__regex'
            elif tokens.get('between'):
                field += '__range'
                value = (tokens['min'], tokens['max'])
            else:
                raise ParseError("unknown clause token {}".format(tokens))

            q = Q(**{field: value})
            if negation:
                q = ~q
            return q, annotations

        def parse_filter(tokens):
            query = Q()
            annotations = {}
            operation = 'and'
            negation = False
            for token in tokens:
                if isinstance(token, p.ParseResults):
                    q, ann = parse_filter(token)
                    query &= q
                    annotations.update(ann)
                elif isinstance(token, str):
                    if token in ('or', 'and'):
                        operation = token
                    elif token == 'not':
                        negation = True
                else:
                    q, ann = token
                    annotations.update(ann)
                    if negation:
                        q = ~q
                    if operation == 'or':
                        query |= q
                    else:
                        query &= q
            return query, annotations

        def parse_model(tokens):
            return self.available_models[tokens['model']]

        def parse(tokens):
            query, annotations = tokens.get('filter', (Q(), {}))
            model = tokens['model']
            model_name = model._meta.model_name
            qs = model._default_manager.annotate(**annotations).filter(query)
            custom_model_name = tokens.get('model_name')
            if custom_model_name:
                model_name = custom_model_name[0]
            result = ParseResult()
            result.queryset = qs
            result.model_name = model_name
            result.aliases = {
                name: field for name, field in tokens.get('aliases', [])
            }
            result.aliases.pop(model_name, None)
            return result

        # The grammar
        # TODO: implement operation args eg. substr(.field, 1, 5)
        comments = p.Suppress(p.ZeroOrMore(p.pythonStyleComment))
        point = p.Literal('.')
        expo = p.CaselessLiteral('e')
        plusorminus = p.Literal('+') | p.Literal('-')
        number = p.Word(p.nums)
        integer = p.Combine(p.Optional(plusorminus) + number)
        floatnumber = p.Combine(
            integer
            + p.Optional(point + p.Optional(number))
            + p.Optional(expo + integer)
        )
        hexnumber = p.Combine(
            p.Optional(plusorminus)
            + p.CaselessLiteral('0x')
            + p.Word(p.hexnums)
        )
        numscalar = (hexnumber | floatnumber | integer).setParseAction(
            lambda t: ast.literal_eval(t[0])
        )
        boolean = (p.Keyword('true') | p.Keyword('false')).setParseAction(
            lambda t: t[0].lower() == 'true'
        )
        alpha_under = p.Regex(r'[a-z][a-z0-9]*(_[a-z0-9]+)*', re.I)
        enumvalue = p.Combine(
            alpha_under + '.' + alpha_under + '.' + alpha_under
        ).setParseAction(parse_enum)
        string = (
            p.Optional(p.Literal('i')).setResultsName('nocase')
            + p.quotedString.setParseAction(p.removeQuotes).setResultsName(
                'string'
            )
        ).setParseAction(parse_string)
        value = (
            (string | numscalar | boolean | enumvalue)
            .setResultsName('value')
            .setParseAction(lambda t: t[0])
        )
        model = (
            p.Regex(r'([A-Z][a-z0-9]*)+')
            .setParseAction(parse_model)
            .setResultsName('model')
        )
        field_name = (
            p.OneOrMore(p.Suppress('.') + alpha_under)
            .setParseAction(parse_field)
            .setResultsName('field')
        )
        operation = (
            p.Word(p.alphas).setResultsName('func_name')
            + p.Suppress('(')
            + field_name.setResultsName('func_field')
            + p.Suppress(')')
        ).setResultsName('operation')
        field = field_name | operation
        is_kw = p.Keyword('is')
        negation = (
            p.Optional(p.Keyword('not'))
            .setParseAction(lambda t: bool(t))
            .setResultsName('negation')
        )
        negation_does = (
            p.Optional(p.Keyword("doesn't") | p.Keyword("does not"))
            .setParseAction(lambda t: bool(t))
            .setResultsName('negation')
        )
        equality = (field + p.Suppress('=') + value).setResultsName('=')
        inequality = (field + p.Suppress('!=') + value).setResultsName('!=')
        lt = (field + p.Suppress('<') + value).setResultsName('<')
        lte = (field + p.Suppress('<=') + value).setResultsName('<=')
        gt = (field + p.Suppress('>') + value).setResultsName('>')
        gte = (field + p.Suppress('>=') + value).setResultsName('>=')
        null_or_none = p.Keyword('null') | p.Keyword('none')
        null = (field + is_kw + negation + null_or_none).setResultsName('null')
        empty = (field + is_kw + negation + p.Keyword('empty')).setResultsName(
            'empty'
        )
        contains = (
            field
            + negation_does
            + (p.Keyword('contain') | p.Keyword('contains'))
            + string.setResultsName('value')
        ).setResultsName('contains')
        startswith = (
            field
            + negation_does
            + (p.Keyword('start with') | p.Keyword('starts with'))
            + string.setResultsName('value')
        ).setResultsName('startswith')
        endswith = (
            field
            + negation_does
            + (p.Keyword('end with') | p.Keyword('ends with'))
            + string.setResultsName('value')
        ).setResultsName('endswith')
        matches = (
            field
            + negation_does
            + (p.Keyword('match') | p.Keyword('matches'))
            + string.setResultsName('value')
        ).setResultsName('matches')
        between = (
            field
            + negation
            + p.Keyword('between')
            + value.setResultsName('min')
            + p.Keyword('and')
            + value.setResultsName('max')
        ).setResultsName('between')
        clause = (
            (
                equality
                | inequality
                | lte
                | gte
                | lt
                | gt
                | between
                | null
                | empty
                | contains
                | startswith
                | endswith
                | matches
            )
            .setParseAction(parse_clause)
            .setResultsName('comparison')
        )
        oper_not = p.Keyword('not')
        oper_and = p.Keyword('and')
        oper_or = p.Keyword('or')
        filter = p.operatorPrecedence(
            clause + comments,
            [
                (oper_not, 1, p.opAssoc.RIGHT),
                (oper_or, 2, p.opAssoc.LEFT),
                (p.Optional(oper_and, default='and'), 2, p.opAssoc.LEFT),
            ],
        ).setParseAction(parse_filter)
        alias = (
            p.Keyword('alias')
            + alpha_under.setResultsName('name')
            + p.Optional(p.Suppress('=') + field_name.setResultsName('field'))
            + comments
        ).setParseAction(lambda t: (t['name'], t.get('field', t['name'])))
        stmt = (
            p.stringStart()
            + comments
            + model
            + p.Optional(
                (p.Suppress(p.Keyword('as')) + alpha_under)
            ).setResultsName('model_name')
            + comments
            + p.Optional(filter).setResultsName('filter')
            + comments
            + p.ZeroOrMore(alias).setResultsName('aliases')
            + comments
            + p.StringEnd()
        ).setParseAction(parse)

        return stmt.parseString(query)[0]


def parse_query(query: str) -> ParseResult:
    return QueryParser().parse_query(query)



import celery

from django.db import transaction

from massmailer.models import MailState, BatchEmail


@celery.shared_task(
    bind=True,
    ignore_result=True,
    default_retry_delay=60,
    max_retries=2,
    soft_time_limit=10,
    time_limit=60,
)
def send_email(self, mail_id):
    def set_state(old_state, new_state):
        with transaction.atomic():
            try:
                email = BatchEmail.objects.get(
                    pk=mail_id, state=old_state.value
                )
            except BatchEmail.DoesNotExist:
                return None
            email.state = new_state.value
            email.save()
        return email

    # atomically push to *sending* queue
    email = set_state(MailState.pending, MailState.sending)
    if not email:
        # deleted or already sent
        return

    python_mail = email.build_email()

    try:
        # this can take a long time or fail
        ret = python_mail.send(fail_silently=False)
        if not ret:
            raise RuntimeError(
                "mail.send() should have returned a truthful value; returned {}".format(
                    ret
                )
            )
    except Exception as exc:
        # atomically push to *pending* queue (for retry)
        if not set_state(MailState.sending, MailState.pending):
            # deleted or already pending or already sent
            return
        raise self.retry(exc=exc)

    # mark as sent
    set_state(MailState.sending, MailState.sent)



REGISTERED_ENUMS = {}


def register_enum(namespace, enum_name=None):
    def decorator(f):
        nonlocal enum_name
        if enum_name is None:
            enum_name = f.__name__
        REGISTERED_ENUMS[namespace + '.' + enum_name] = f
        return f

    return decorator



import copy
import re

from crispy_forms.bootstrap import StrictButton
from crispy_forms.helper import FormHelper
from django import forms
from django.conf import settings
from django.core.validators import RegexValidator
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import get_language, ugettext_lazy as _
from reversion.models import Version

import massmailer.models


class TemplateForm(forms.ModelForm):
    html_enabled = forms.BooleanField(
        initial=False, required=False, label=_("HTML enabled")
    )
    use_markdown = forms.BooleanField(
        initial=True, required=False, label=_("Generate HTML from plaintext")
    )
    useful_queries = forms.ModelChoiceField(
        queryset=massmailer.models.Query.objects.all(),
        required=False,
        label=_("Useful queries"),
    )
    # for overwrite checking
    revisions = forms.IntegerField(widget=forms.HiddenInput())

    class Meta:
        model = massmailer.models.Template
        fields = [
            'name',
            'description',
            'subject',
            'plain_body',
            'html_body',
            'language',
        ]
        widgets = {'description': forms.Textarea(attrs={'rows': 2})}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['revisions'].initial = self._revision_count()
        if (
            self.instance
            and self.instance.pk
            and self.instance.html_body.strip()
        ):
            self.fields['html_enabled'].initial = True

        self.fields['language'] = forms.ChoiceField(
            choices=settings.LANGUAGES,
            initial=get_language(),
            widget=forms.Select(),
            required=True,
        )

    def _revision_count(self):
        if self.instance and self.instance.pk:
            return Version.objects.get_for_object(self.instance).count()
        return 0

    def reset_overwritten(self):
        # FIXME: hack
        self.data._mutable = True
        self.data['revisions'] = self._revision_count()
        self.data._mutable = False

    def was_overwritten(self):
        return self.cleaned_data['revisions'] != self._revision_count()

    def clean(self):
        data = self.cleaned_data
        if data['html_enabled'] and not data.get('html_body', '').strip():
            raise forms.ValidationError(
                _("You must provide an HTML template if HTML is enabled.")
            )


class QueryForm(forms.ModelForm):
    class Meta:
        model = massmailer.models.Query
        fields = ['name', 'description', 'useful_with', 'query']
        widgets = {'description': forms.Textarea(attrs={'rows': 2})}


class CreateBatchForm(forms.ModelForm):
    FOOLPROOF_PHRASE = _("i am fine bothering %(n)s people")

    class Meta:
        model = massmailer.models.Batch
        fields = ('name', 'template', 'query')

    @classmethod
    def foolproof_field(cls, count):
        phrase = cls.FOOLPROOF_PHRASE % {'n': count}
        field = forms.CharField(
            label=_("Foolproofing"),
            widget=forms.TextInput(
                attrs={'autocomplete': 'off', 'autofocus': True}
            ),
        )
        field.help_text = _("Type “%s” above") % "\ufeff".join(phrase)
        field.validators.append(RegexValidator(re.escape(phrase)))
        return field

    def foolproof_enabled(self):
        return not settings.DEBUG

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = self.data.copy()

        submit = _("Send")
        submit_cls = "btn-primary"

        if 'query' in self.data and 'template' in self.data:
            # once the form is submitted once, add the foolproof test (we now know the user count)
            query = self.fields['query'].queryset.get(pk=self.data['query'])

            result, user_qs = massmailer.models.Query.execute(query.query)
            count = len(user_qs)
            submit = _("Actually send to %(n)s people right now") % {
                'n': count
            }
            submit_cls = "btn-warning"

            for field in CreateBatchForm.Meta.fields:
                f = self.fields[field]
                name = 'dis_' + field
                disabled_copy = copy.deepcopy(f)  # add a disabled copy
                disabled_copy.disabled = True
                self.fields[name] = disabled_copy
                self.initial[name] = self.data[field]  # copy value
                f.widget = forms.HiddenInput()  # hide original data

            if not self.data.get('name'):
                # generate a nice name
                template = self.fields['template'].queryset.get(
                    pk=self.data['template']
                )
                name = '{} ⋅ {} ⋅ {}'.format(
                    template, query, timezone.now().date()
                )
                self.data['name'] = self.initial['name'] = self.initial[
                    'dis_name'
                ] = name

            if self.foolproof_enabled():
                # add foolproof
                self.fields['foolproof'] = self.foolproof_field(count)

        self.helper = FormHelper(self)
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-lg-4'
        self.helper.field_class = 'col-lg-8'
        # add the submit button
        self.helper.layout.append(
            StrictButton(
                format_html('<i class="fa fa-paper-plane-o"></i> {}', submit),
                type="submit",
                css_class="{} btn-block".format(submit_cls),
            )
        )



from django.urls import path, include

import massmailer.views

app_name = 'massmailer'

template_patterns = [
    path('', massmailer.views.CreateTemplateView.as_view(), name='new'),
    path(
        'preview',
        massmailer.views.TemplatePreviewView.as_view(),
        name='preview',
    ),
    path(
        '<int:id>',
        massmailer.views.UpdateTemplateView.as_view(),
        name='update',
    ),
    path(
        '<int:id>-<slug>',
        massmailer.views.UpdateTemplateView.as_view(),
        name='update',
    ),
]

query_patterns = [
    path('', massmailer.views.CreateQueryView.as_view(), name='new'),
    path(
        'preview', massmailer.views.QueryPreviewView.as_view(), name='preview'
    ),
    path(
        '<int:id>', massmailer.views.UpdateQueryView.as_view(), name='update'
    ),
    path(
        '<int:id>-<slug>',
        massmailer.views.UpdateQueryView.as_view(),
        name='update',
    ),
]

batch_obj_patterns = [
    path('', massmailer.views.BatchDetailView.as_view(), name='detail'),
    path(
        'retry',
        massmailer.views.BatchRetryView.as_view(),
        name='retry-pending',
    ),
    path('delete', massmailer.views.BatchDeleteView.as_view(), name='delete'),
]

batch_patterns = [
    path('', massmailer.views.BatchListView.as_view(), name='list'),
    path('new', massmailer.views.BatchCreateView.as_view(), name='new'),
    path('<int:id>/', include(batch_obj_patterns)),
]

urlpatterns = [
    path('', massmailer.views.DashboardView.as_view(), name='dashboard'),
    path(
        'template/',
        include((template_patterns, app_name), namespace='template'),
    ),
    path('query/', include((query_patterns, app_name), namespace='query')),
    path('batch/', include((batch_patterns, app_name), namespace='batch')),
]



import os.path
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as fh:
    long_description = fh.read()

setup(
    name="django-massmailer",
    version='0.2',
    author="Association Prologin",
    author_email="info@prologin.org",
    license="GPL3",
    packages=["massmailer"],
    include_package_data=True,
    description=(
        "A standalone Django app to send templated emails in batch. "
        "Features a custom query engine and template editor with preview."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "babel",  # i18n template filters
        "bleach",  # HTML sanitizer
        "celery>4",  # task queue
        "django-bootstrap-breadcrumbs",  # breadcrumbs in default templates
        "django-crispy-forms",  # form builder compatible with Bootstrap
        "django-reversion",  # model revision/history
        "django>=2.1",
        "jinja2",  # sane templates
        "markdown",
        "pyparsing",  # query language parser
    ],
    classifiers=[
        'Environment :: Web Environment',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'Topic :: Communications :: Email',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    project_urls={
        'Source': 'https://github.com/prologin/django-massmailer/',
        'Issue Tracker': 'https://github.com/prologin/django-massmailer/issues',
    },
)



import bleach
import enum
import jinja2
import jinja2.meta
import jinja2.runtime
import jinja2.sandbox
import operator
import re
import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import mail
from django.core.exceptions import FieldDoesNotExist
from django.urls import reverse
from django.db import models
from django.db.models import Count, F, BooleanField
from django.utils import timezone
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _
from functools import reduce

from massmailer.query_parser import QueryParser, ParseError
from massmailer.utils.db import ConditionalSum, CaseMapping
from massmailer.utils import filters as mfilters

TEMPLATE_OPTS = {
    'autoescape': False,
    'trim_blocks': True,
    'undefined': jinja2.runtime.StrictUndefined,
}

VARIABLE_PLACEHOLDER = '<span class="placeholder">\u25cc</span>'
RE_TAG = re.compile(
    r'\{([%#])(.*?)\1\}|\{\{(.*?)\}\}', re.MULTILINE | re.DOTALL
)


class MailState(enum.IntEnum):
    pending = 1
    sending = 2
    sent = 3
    delivered = 4
    bounced = 5
    complained = 6

    @classmethod
    def bad(cls):
        return {cls.bounced, cls.complained}


class TemplateItem(enum.Enum):
    subject = ('subject', {})
    plain = ('plain_body', {})
    html = ('html_body', {'autoescape': True})


class Template(models.Model):
    name = models.CharField(max_length=144, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    subject = models.TextField(verbose_name=_("Subject template"))
    plain_body = models.TextField(verbose_name=_("Plaintext body template"))
    html_body = models.TextField(
        blank=True, verbose_name=_("HTML body template")
    )
    language = models.CharField(max_length=10, verbose_name=_("Language"))

    class Meta:
        ordering = ['name']
        verbose_name = _("Template")
        verbose_name_plural = _("Templates")

    def __str__(self):
        return self.name

    @staticmethod
    def template_opts(item: TemplateItem):
        _, specific_opts = item.value
        opts = TEMPLATE_OPTS.copy()
        opts.update(specific_opts)
        return opts

    @property
    def html_enabled(self):
        return bool(self.html_body.strip())

    def template_source(self, item: TemplateItem):
        attr, _ = item.value
        return getattr(self, attr)

    def environment(self, item):
        env = jinja2.sandbox.SandboxedEnvironment(**self.template_opts(item))
        env.filters['format_datetime'] = mfilters.format_datetime
        env.filters['format_date'] = mfilters.format_date
        env.filters['format_time'] = mfilters.format_time
        return env

    def template(self, item: TemplateItem):
        env = self.environment(item)
        return env.from_string(source=self.template_source(item))

    def variables(self, item: TemplateItem):
        env = self.environment(item)
        ast = env.parse(source=self.template_source(item))
        return jinja2.meta.find_undeclared_variables(ast)

    def render(self, item: TemplateItem, context: dict):
        context['language'] = self.language
        content = self.template(item).render(context)
        if item is TemplateItem.html:
            content = bleach.linkify(content)
        return content

    def preview(self, item: TemplateItem):
        text = self.template_source(item)

        def replace(match):
            return jinja2.Markup(VARIABLE_PLACEHOLDER)

        return RE_TAG.sub(replace, text)

    def full_preview(self, context):
        result = {}
        for item in TemplateItem:
            data = {}
            try:
                declared = self.variables(item)
                rendered = self.render(item, context)
                data['context'] = {
                    'declared': list(context.keys() - declared),
                    'missing': list(declared - context.keys()),
                }
                data['content'] = rendered
            except jinja2.UndefinedError as error:
                data['error'] = {'type': 'undefined', 'msg': error.message}
            except jinja2.TemplateSyntaxError as error:
                data['error'] = {'type': 'syntax', 'msg': error.message}
            except Exception as error:
                data['error'] = {'type': 'other', 'msg': str(error)}
            result[item.name] = data
        return result

    def subject_preview(self):
        return self.preview(TemplateItem.subject)

    def plain_preview(self):
        return self.preview(TemplateItem.plain)

    def html_preview(self):
        return self.preview(TemplateItem.html)

    def get_absolute_url(self):
        return reverse(
            'massmailer:template:update',
            kwargs={'id': self.pk, 'slug': slugify(self.name)},
        )


class Query(models.Model):
    name = models.CharField(max_length=144, verbose_name=_("Name"))
    description = models.TextField(blank=True, verbose_name=_("Description"))
    query = models.TextField(verbose_name=_("Query"))
    useful_with = models.ManyToManyField(
        Template,
        blank=True,
        related_name='useful_queries',
        verbose_name=_("Useful with templates"),
    )

    class Meta:
        ordering = ['name']
        verbose_name = _("Query")
        verbose_name_plural = _("Queries")

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse(
            'massmailer:query:update',
            kwargs={'id': self.pk, 'slug': slugify(self.name)},
        )

    def get_results(self):
        return self.execute(self.query)

    def parse(self):
        return QueryParser().parse_query(self.query)

    @staticmethod
    def execute(query):
        User = get_user_model()
        user_label = User._meta.label

        result = QueryParser().parse_query(query)
        qs = result.queryset
        qs_label = qs.model._meta.label

        user_qs = qs
        # if queryset is not a User queryset, require a "user" alias
        if user_qs.model != User:
            if 'user' not in result.aliases:
                raise ParseError(
                    _(
                        "The root model is not %(model)s. You must provide a `user` alias."
                    )
                    % {'model': user_label}
                )
            user_field = result.aliases['user']
            try:
                if qs.model._meta.get_field(user_field).related_model != User:
                    raise ParseError(
                        _("%(label)s.%(field)s is not %(model)s")
                        % {
                            'label': qs_label,
                            'field': user_field,
                            'model': user_label,
                        }
                    )
            except FieldDoesNotExist:
                raise ParseError(
                    _(
                        "%(label)s has no field `%(field)s`"
                        % {'label': qs_label, 'field': user_field}
                    )
                )
            user_pks = set(qs.values_list(user_field, flat=True))
            user_qs = User._default_manager.filter(pk__in=user_pks)

        return result, user_qs


class BatchManager(models.Manager):
    def get_queryset(self):
        total = F('email_count')
        ecount = lambda e: F('{}_email_count'.format(e))

        qs = (
            super()
            .get_queryset()
            .select_related('query', 'template', 'initiator')
            .prefetch_related('emails')
        )

        def annotate(key, value):
            nonlocal qs
            qs = qs.annotate(**{key: value})

        def annotate_count(name, expr):
            count_field = ecount(name)
            annotate(count_field.name, expr)
            annotate('{}_percentage'.format(name), 100.0 * count_field / total)

        annotate(total.name, Count('emails'))

        for state in MailState:
            annotate_count(
                state.name, ConditionalSum(emails__state=state.value)
            )

        annotate_count('unsent', ecount('pending') + ecount('sending'))
        annotate_count(
            'erroneous',
            reduce(
                operator.add, (ecount(state.name) for state in MailState.bad())
            ),
        )
        annotate(
            'completed',
            CaseMapping(
                ecount('unsent').name,
                [(0, True)],
                default=False,
                output_field=BooleanField(),
            ),
        )
        return qs


class Batch(models.Model):
    name = models.CharField(
        max_length=140, blank=True, verbose_name=_("Optional name")
    )
    template = models.ForeignKey(
        Template,
        null=True,
        on_delete=models.SET_NULL,
        related_name='batches',
        verbose_name=_("Template"),
    )
    query = models.ForeignKey(
        Query,
        null=True,
        on_delete=models.SET_NULL,
        related_name='batches',
        verbose_name=_("Query"),
    )
    initiator = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        on_delete=models.SET_NULL,
        related_name='massmailer_batches',
    )
    date_created = models.DateTimeField(default=timezone.now, null=False)

    class Meta:
        ordering = ['-date_created']
        verbose_name = _("Batch")
        verbose_name_plural = _("Batches")

    objects = BatchManager()

    @property
    def default_name(self):
        return _("Batch %(id)s") % {'id': self.pk}

    def __str__(self):
        return self.name or self.default_name

    def pending_emails(self):
        return self.emails.filter(state=MailState.pending.value)

    @property
    def erroneous_emails(self):
        return self.emails.filter(
            state__in=[state.value for state in MailState.bad()]
        )

    def build_emails(self):
        result, user_qs = self.query.get_results()
        queryset = result.queryset.order_by('pk')

        if result.queryset.model is get_user_model():
            user_getter = lambda object: object
        else:
            user_getter = lambda object: getattr(
                object, result.aliases['user']
            )

        html_enabled = self.template.html_enabled

        for object in queryset:
            context = {
                alias: getattr(object, field)
                for alias, field in result.aliases.items()
            }
            context[result.model_name] = object
            user = user_getter(object)
            yield BatchEmail(
                batch=self,
                user=user,
                to=user.email,
                subject=self.template.render(TemplateItem.subject, context),
                body=self.template.render(TemplateItem.plain, context),
                html_body=self.template.render(TemplateItem.html, context)
                if html_enabled
                else "",
            )


class BatchEmail(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    state = models.PositiveIntegerField(
        db_index=True, default=MailState.pending.value
    )
    batch = models.ForeignKey(
        Batch, related_name='emails', on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL
    )

    to = models.EmailField(
        blank=False
    )  # in case user is deleted or changes email
    subject = models.TextField(blank=True)
    body = models.TextField(blank=True)
    html_body = models.TextField(blank=True, default="")

    class Meta:
        ordering = ['state']

    def __str__(self):
        return '[{}] {}'.format(self.state_display, self.id)

    @property
    def state_display(self):
        return MailState(self.state).name

    @property
    def task_id(self):
        return 'massmailer-{}'.format(self.pk)

    @property
    def pending(self):
        return self.state == MailState.pending.value

    def build_email(self):
        assert self.subject
        assert self.body
        # add a custom header to resolve the mail ID from bounces/complaints
        headers = {'X-MID': self.id}
        if self.user and hasattr(self.user, 'get_unsubscribe_url'):
            headers['List-Unsubscribe'] = '<{}>'.format(
                self.user.get_unsubscribe_url()
            )

        kwargs = {
            'to': [self.to],
            'subject': self.subject,
            'body': self.body,
            'headers': headers,
        }

        if self.html_body:
            email = mail.EmailMultiAlternatives(**kwargs)
            email.attach_alternative(self.html_body, 'text/html')
        else:
            email = mail.EmailMessage(**kwargs)
        return email

    def send_task(self):
        from massmailer.tasks import send_email

        return send_email.apply_async(args=[self.pk], task_id=self.task_id)



from reversion.admin import VersionAdmin

from django.contrib import admin
from django.db.models import Count
from django.utils.translation import ugettext_lazy as _

import massmailer.models


class TemplateAdmin(VersionAdmin):
    exclude = ['author']


class QueryAdmin(VersionAdmin):
    exclude = ['author', 'useful_with']


class BatchEmailInline(admin.TabularInline):
    model = massmailer.models.BatchEmail
    fields = readonly_fields = ['id', 'user', 'to', 'state_display']
    extra = max_num = 0
    can_delete = False


class BatchAdmin(admin.ModelAdmin):
    inlines = [BatchEmailInline]
    raw_id_fields = ['initiator']
    search_fields = ['initiator__username', 'template__name', 'query__name']
    list_display = ['__str__', 'initiator', 'date_created', 'email_count']
    list_filter = ['template', 'query']

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related('initiator')
            .prefetch_related('emails')
            .annotate(email_count=Count('emails'))
        )

    def email_count(self, obj):
        return obj.email_count

    email_count.short_description = _("Email count")
    email_count.admin_order_field = 'email_count'


admin.site.register(massmailer.models.Template, TemplateAdmin)
admin.site.register(massmailer.models.Query, QueryAdmin)
admin.site.register(massmailer.models.Batch, BatchAdmin)
